# This module is automatically created from setup.py
git_sha1 = None
local_modifications = None
version = '2023.06.07'
long_version = 'pyrocko-eost-2023.06.07-2023-08-14_14:18:21'  # noqa
installed_date = '2023-08-14_14:18:21'
src_path = 'C:\\Users\\kylia\\Desktop\\pyrocko_test'
